"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Container = void 0;
const types_1 = require("./types");
const Definition_1 = require("./Definition");
const assertNoCircularDependencies_1 = require("./assertNoCircularDependencies");
const ContainerArg_1 = require("./args/ContainerArg");
const is = require("predicates");
const randomName_1 = require("./randomName");
const debugFn_1 = require("./debugFn");
const errors_1 = require("./errors");
const debugCreation = (0, debugFn_1.debugFn)("creation");
const debugDefinition = (0, debugFn_1.debugFn)("definition");
const debugSlowCreation = (0, debugFn_1.debugFn)("slow-creation");
class Container {
    constructor(parent) {
        this.definitions = new Map();
        this.services = new Map();
        this.middlewares = [];
        /**
         * Time needed to trigger debug slow creation log
         */
        this.slowLogThreshold = 10000;
        Object.defineProperty(this, "parent", {
            value: parent,
            writable: false,
        });
    }
    /**
     * Registers given service definition
     */
    registerDefinition(definition) {
        if (this.definitions.has(definition.name)) {
            throw errors_1.ERRORS.ALREADY_DEFINED.create(definition.name.toString());
        }
        debugDefinition(`Service ${definition.name.toString()} defined`);
        definition.setOwner(this);
        this.definitions.set(definition.name, definition);
        return this;
    }
    /**
     * Creates and registers service definition
     *
     * Returns created definition for further configuration
     */
    definition(name) {
        const definition = new Definition_1.Definition(name);
        this.registerDefinition(definition);
        return definition;
    }
    definitionWithConstructor(nameOrClazz, clazz) {
        const finalClazz = is.func(nameOrClazz) ? nameOrClazz : clazz;
        return this.definition(types_1.ServiceName.is(nameOrClazz) ? nameOrClazz : (0, randomName_1.randomName)(finalClazz.name)).useConstructor(finalClazz);
    }
    definitionWithFactory(nameOrFactory, factoryOrType, type) {
        const name = types_1.ServiceName.is(nameOrFactory) ? nameOrFactory : undefined;
        const factory = is.func(nameOrFactory)
            ? nameOrFactory
            : factoryOrType;
        const finalType = is.func(nameOrFactory) ? factoryOrType : type;
        const def = this.definition(name).useFactory(factory);
        if (finalType) {
            def.markType(finalType);
        }
        return def;
    }
    definitionWithValue(nameOrValue, value) {
        const isNameProvided = value !== undefined && types_1.ServiceName.is(nameOrValue);
        const finalValue = isNameProvided ? value : nameOrValue;
        const name = isNameProvided
            ? nameOrValue
            : (0, randomName_1.randomName)(Object.getPrototypeOf(finalValue).constructor.name);
        return this.definition(name).useValue(finalValue);
    }
    /**
     * Returns definition by given name
     */
    findByName(name) {
        if (this.definitions.has(name)) {
            return this.definitions.get(name);
        }
        if (this.parent) {
            return this.parent.findByName(name);
        }
    }
    /**
     * Returns definitions that satisfy given predicate
     */
    findByPredicate(predicate) {
        return Array.from(this.definitions.values())
            .filter(predicate)
            .concat(this.parent ? this.parent.findByPredicate(predicate) : []);
    }
    findByAnnotation(predicate, withAnnotation = false) {
        let annotations = [];
        const definitions = this.findByPredicate(s => {
            const annotation = s.annotations.find(predicate);
            if (annotation) {
                withAnnotation && annotations.push(annotation);
                return true;
            }
            return false;
        });
        if (withAnnotation) {
            return definitions.map((d) => {
                return [d, annotations.shift()];
            });
        }
        return definitions;
    }
    /**
     * Registers given middleware
     */
    addMiddleware(...middlewares) {
        for (const middleware of middlewares) {
            if (middleware[types_1.onMiddlewareAttach]) {
                middleware[types_1.onMiddlewareAttach](this);
            }
        }
        this.middlewares.push(...middlewares);
        return this;
    }
    getMiddlewares() {
        return this.middlewares;
    }
    /**
     * Returns service for given name or definition
     */
    get(nameOrDefinition) {
        let definition;
        if (types_1.ServiceName.is(nameOrDefinition)) {
            definition = this.findByName(nameOrDefinition);
            if (definition === undefined) {
                return Promise.reject(errors_1.ERRORS.SERVICE_NOT_FOUND.create(nameOrDefinition.toString()));
            }
        }
        else {
            definition = nameOrDefinition;
        }
        if (this.services.has(definition)) {
            return this.services.get(definition);
        }
        if (!this.definitions.has(definition.name) && this.parent) {
            return this.parent.get(definition);
        }
        const promise = this.create(definition);
        this.services.set(definition, promise);
        return promise;
    }
    async create(definition) {
        if (!definition.factory) {
            return Promise.reject(errors_1.ERRORS.INCOMPLETE_DEFINITION.create(`Missing factory for service definition "${definition.name.toString()}". Define it as constructor, factory or value`));
        }
        (0, assertNoCircularDependencies_1.assertNoCircularDependencies)(this, definition);
        // valid definition, time to lock it
        definition.lock();
        let timeout;
        if (this.slowLogThreshold > 0) {
            timeout = setTimeout(() => {
                debugSlowCreation(`Service ${definition.name.toString()} takes a long time to create. Over ${this.slowLogThreshold} ms`);
            }, this.slowLogThreshold);
        }
        const debugMsg = `Creating service ${definition.name.toString()}`;
        debugCreation(`${debugMsg} - started`);
        let currentMiddleware = 0;
        const middlewares = this.getMiddlewares();
        const next = (definition) => {
            const middleware = middlewares[currentMiddleware];
            currentMiddleware++;
            if (middleware) {
                return middleware.call(this, definition, next);
            }
            return Promise.all(definition.args.map(a => (ContainerArg_1.ContainerArg.is(a) ? a.getArgument(this) : a))).then((args) => definition.factory.apply(this, args));
        };
        try {
            const result = await next.call(this, definition);
            timeout && clearTimeout(timeout);
            debugCreation(`${debugMsg} - finished`);
            return result;
        }
        catch (e) {
            timeout && clearTimeout(timeout);
            throw e;
        }
    }
    /**
     * Returns all services that definition satisfies predicate
     */
    getByPredicate(predicate) {
        return Promise.all(this.findByPredicate(predicate).map(d => this.get(d)));
    }
    getByAnnotation(predicate, withAnnotation = false) {
        return Promise.all(this.findByAnnotation(predicate, true).map(async ([definition, annotation]) => {
            if (withAnnotation) {
                return [await this.get(definition), annotation];
            }
            return this.get(definition);
        }));
    }
    alias(definition, options) {
        const newDefinition = definition.createAlias(options);
        this.registerDefinition(newDefinition);
        return newDefinition;
    }
}
exports.Container = Container;
